@@ -1,14 +1,17 @@
 
 import sys
+import warnings
 from functools import partial
 import os
 import shutil
-import csv
 import yaml
 import subprocess
 from argparse import ArgumentParser, Action
 import deft.tracker
-from deft.tracker import UserError
+from deft.warn import PrintWarnings
+from deft.tracker import UserError, FormatVersion
+from deft.upgrade import create_upgrader
+from deft.formats import *
 
 
 EditorEnvironmentVariables = ["DEFT_EDITOR", "VISUAL", "EDITOR"]
@@ -30,7 +33,7 @@ def run_editor_process(path):
 
 def with_tracker(function):
     def wrapper(self, args):
-        tracker = self.backend.load_tracker()
+        tracker = self.backend.load_tracker(self.warning_output)
         function(self, tracker, args)
         tracker.save()
     
@@ -41,9 +44,6 @@ def _ignore_output(s):
     pass
     
 
-
-
-
 def append_value(namespace, name, value):
     values = getattr(namespace, name) or []
     values = values + [value]
@@ -64,12 +64,28 @@ class AppendPropertyDeleter(Action):
         append_value(namespace, self.dest, deleter)
 
 
+_DuplicateEntriesMessage = \
+    "found multiple index entries for feature {feature.name}: " + \
+    "ignored duplicate entry in index of status {removed_from_status}, " + \
+    "using status {feature.status}, priority {feature.priority}"
+
+_UnindexedFeatureMessage = \
+    "feature {feature.name} did not have a status or priority: " + \
+    "assigned status {feature.status}, requires manual repair"
+
+_UnknownFeatureMessage = \
+    "nonexistent feature {name} found in index for status {status}: removed from index"
+                               
+
 class CommandLineInterface(object):
     def __init__(self, backend, out, err, editor=run_editor_process):
         self.backend = backend
         self.editor = editor
         self.out = out
-        self.err = err
+        self.warning_output = PrintWarnings(err, "WARNING: ",
+            duplicate_entries=_DuplicateEntriesMessage,
+            unindexed_feature=_UnindexedFeatureMessage,
+            unknown_feature=_UnknownFeatureMessage)
         
     def run(self, argv):
         command = argv[0]
@@ -150,17 +166,24 @@ class CommandLineInterface(object):
                                  help="output in CSV format (default is human-readable text)",
                                  dest="format",
                                  action="store_const",
-                                 const="csv",
-                                 default="text")
+                                 const=write_table_as_csv,
+                                 default=write_table_as_text)
         
         statuses_parser = subparsers.add_parser("statuses",
                                                 help="list all active statuses")
         
         status_parser = subparsers.add_parser("status", 
                                               help="query or change the status of a feature")
-        status_parser.add_argument("feature",
-                                   help="feature name",
+        status_parser.add_argument("name",
+                                   help="feature name, or status name if the --all-with-status flag is set",
                                    metavar="name")
+        status_parser.add_argument("-s", "--all-with-status",
+                                   help="change the status of all features that currently have status S",
+                                   dest="name_is_status",
+                                   action="store_const",
+                                   const=True,
+                                   default=False)
+                                   
         status_parser.add_argument("status",
                                    help="the new status of the feature, if changing the status",
                                    nargs="?",
@@ -266,7 +289,7 @@ class CommandLineInterface(object):
         if args.initial_status is not None:
             config['initial_status'] = args.initial_status
         
-        self.backend.init_tracker(**config)
+        self.backend.init_tracker(self.warning_output, **config)
         args.info_output("initialised Deft tracker")
     
     
@@ -308,20 +331,22 @@ class CommandLineInterface(object):
         
         table = features_to_table(features, args.properties)
         
-        if args.format == "text":
-            write_features_as_text(table, self.out)
-        elif args.format == "csv":
-            write_features_as_csv(table, self.out)
-        else:
-            raise ValueError("unexpected format: " + args.format)
+        args.format(table, self.out)
     
     @with_tracker
     def run_status(self, tracker, args):
-        feature = tracker.feature_named(args.feature)
-        if args.status is not None:
-            feature.status = args.status
+        if args.name_is_status:
+            if args.status is None:
+                raise UserError("new status not specified")
+            else:
+                tracker.bulk_change_status(from_status=args.name, to_status=args.status)
         else:
-            self.println(feature.status)
+            feature = tracker.feature_named(args.name)
+            if args.status is None:
+                self.println(feature.status)
+            else:
+                feature.status = args.status
+
     
     @with_tracker
     def run_priority(self, tracker, args):
@@ -380,9 +405,12 @@ class CommandLineInterface(object):
             tracker.purge(name)
     
     def run_upgrade_format(self, args):
-        from deft.upgrade import upgrade
-        upgrade(self.backend.tracker_storage())
-        
+        upgrader = create_upgrader()
+        storage = self.backend.tracker_storage()
+        if upgrader.upgrade(storage):
+            args.info_output("upgraded to format version " + FormatVersion)
+        else:
+            args.info_output("already at format version " + FormatVersion)
     
     def println(self, text):
         self.out.write(text)
@@ -400,39 +428,13 @@ def features_to_table(features, property_names=[]):
     return [tuple([f.status, f.priority, f.name] + property_values(f, property_names)) for f in features]
 
 
-def write_features_as_text(features_table, out):
-    if not features_table:
-        return
-    
-    max_elts = partial(map, max)
-    
-    alignl = "{1:<{0}}".format
-    alignr = "{1:>{0}}".format
-    def align_for(v):
-        return alignr if type(v) == int else alignl
-
-    table = [map(str,t) for t in features_table]
-    zeros = [0 for elt in features_table[0]]
-    aligns = [align_for(v) for v in features_table[0]]
-    col_widths = reduce(max_elts, [map(len,t) for t in table], zeros)
-    col_formatters = map(partial, aligns, col_widths)
-    formatted_table = [[col_formatters[i](row[i]) for i in range(len(row))] for row in table]
-    lines = [" ".join(row) for row in formatted_table]
-    
-    for line in lines:
-        out.write(line)
-        out.write(os.linesep)
-
-
-def write_features_as_csv(features_table, out):
-    csv_out = csv.writer(out)
-    csv_out.writerows(features_table)
-
-
 def main():
     try:
-        CommandLineInterface(deft.tracker, sys.stdout, sys.stderr).run(sys.argv)
+        cli = CommandLineInterface(deft.tracker, sys.stdout, sys.stderr)
+        cli.run(sys.argv)
     except deft.tracker.UserError as e:
         sys.stderr.write(str(e) + "\n")
         sys.exit(1)
+    except KeyboardInterrupt:
+        pass
     # Unexpected exceptions pass through and Python interpreter will print the stacktrace
